// ################################################################################
// Common Framework for Computer Graphics Courses at FI MUNI.
// 
// Copyright (c) 2021-2022 Visitlab (https://visitlab.fi.muni.cz)
// All rights reserved.
// ################################################################################

#include "application.hpp"

// ----------------------------------------------------------------------------
// Constructors & Destructors
// ----------------------------------------------------------------------------

Application::Application(int initial_width, int initial_height, std::vector<std::string> arguments)
    : PV112Application(initial_width, initial_height, arguments) {

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

    glEnable(GL_DEPTH_TEST);

    glEnable(GL_MULTISAMPLE);

    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    compile_shaders();
}

Application::~Application() { Application::delete_shaders(); }

// ----------------------------------------------------------------------------
// Methods
// ----------------------------------------------------------------------------

void Application::delete_shaders() {
    glDeleteProgram(main_program);
}

void Application::compile_shaders() {
    delete_shaders();

    main_program = create_program(lecture_shaders_path / "main.vert", lecture_shaders_path / "main.frag");

    std::cout << "Shaders are reloaded." << std::endl;
}

void Application::update(float delta) {}

void Application::render() {
    glViewport(0, 0, width, height);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glUseProgram(main_program);

    const float cam_distance = 8.0f;
    glm::vec3 camera_position = glm::vec3(
        cam_distance * sin(current_xz_plane_view_angle),
        2.0f,
        cam_distance * cos(current_xz_plane_view_angle)
    );

    glProgramUniform3f(main_program, 6, camera_position.x, camera_position.y, camera_position.z);

    glm::mat4 view_matrix       = glm::lookAt(camera_position, glm::vec3(0.f,0.f,0.f), glm::vec3(0.f, 1.0f, 0.f) );
    glm::mat4 projection_matrix = glm::perspective( glm::radians(45.0f), (float)width/float(height), 0.1f, 100.f );

    //parametry svetla: jsou v runtime nemenne,
    //takze je muzeme plnit takto rovnou "na tvrdo"
    //
    //(to ale znamena, ze vzajemna pozice mezi svetlem a objekty je nemenna
    // a tim padem se nemeni ani nasvicene strany tech objektu; mysi potom
    // posouvame kameru -- cili misto odkud se na situaci divame, svetlo je
    // potom zdanlive spatne protoze nesleduje kameru, ale "putuje s objekty",
    // mys totiz ovlada (jenom) kameru-pozorovatele -- jina varianta by byla,
    // kdyby mys otacela objekty, ale neotacela by kamerou -- tehdy se pozice
    // svetla a objektu meni a svetlo se potom skutecne posouva po
    // objektech/meni se na objektech; nebo se to emuluje tak, ze se svetlo
    // jakoby "prichyti" na kameru, jako celovka na helme hornika)
    //
    //varianta: svetlo fixne ve scene
    //glProgramUniform3f(main_program, 7, -2.f, 3.f, +2.f); //light_position
    //
    //varianta: svetlo fixne ke kamere (tedy se posouva vuci scene)
    const float halfPI = 0; //3.14159f / 2.;
    glProgramUniform3f(main_program, 7,
        cam_distance * sin(current_xz_plane_view_angle+halfPI), //svetlo sviti zpraveho boku
        2.0f,
        cam_distance * cos(current_xz_plane_view_angle+halfPI) ); //light_position
    glProgramUniform3f(main_program, 8, 0.7f, 0.7f, 0.7f); //light_ambient
    glProgramUniform3f(main_program, 9, 1.f, 1.f, 1.f); //_diffuse
    glProgramUniform3f(main_program, 10, 1.f, 1.f, 1.f); //_specular

    // TIPS: Always use f after floating numbers, e.g 1.0f. This helps in avoiding glm template errors
    // Task 5.1: Setup the camera.
    //           1. Choose camera position (for example 1.5f, 2.0f, -8.0f is good to see all our objects) as glm::vec3.
    //           2. Create perspective matrix using glm::perspective(float field_of_view, float aspect_ratio, float near, float far).
    //              - glm works in radians by default, you can convert degrees to radians using glm::radians(degrees)
    //              - aspect_ratio is width of the window divided by height of the window
    //              - near and far planes should be: positive, near < far, near should be much smaller than 1.0f, e.g 0.01f
    //           3. Create view matrix using glm::lookAt(glm::vec3 eye_position, glm::vec3 look_at_position, glm::vec3 up_vector).
    //           4. Leave model matrix as identity for now.
    // Task 5.1: Upload MVP matrices to 'program' - use glm::value_ptr to get pointers for your matrices.

    // Task 5.2: Set uniform data for light and upload camera(eye) position to 'program'.
    // Task 5.2: Set uniform data for material. Remember: it's ambient, diffuse, specular, and additionally shininess for material.

    glProgramUniformMatrix4fv(main_program, 4, 1, false, value_ptr(view_matrix));
    glProgramUniformMatrix4fv(main_program, 5, 1, false, value_ptr(projection_matrix));

    // Task 5.1: Separate each object by moving them to different locations using glm::translate(glm::mat4 matrix_to_translate, glm::vec3).
    //           Play with other glm functions of the same form:
    //            - rotate (mat4x4 matrix_to_rotate, float angle, glm::vec3 axis_to_rotate_around)
    //            - translate (mat4x4 matrix_to_translate, glm::vec3 move_by_vector)
    glm::mat4 model_matrix = glm::translate(glm::mat4(0.1f), glm::vec3(0.f, 0.f, 0.f));
    glProgramUniformMatrix4fv(main_program, 3, 1, false, value_ptr(model_matrix));
    //                                      |
    //                                      +---- location

    //aktualni barvy pro prave renderovany model
    glProgramUniform3f(main_program, 11, 0.0f, 0.2f, 0.0f); //model_ambient
    glProgramUniform3f(main_program, 12, 0.0f, 0.7f, 0.0f); //_diffuse
    glProgramUniform3f(main_program, 13, 1.f, 1.f, 1.f); //_specular
    glProgramUniform1f(main_program, 14, 0.0f); //shinines
    cube.draw();

    model_matrix = glm::translate(glm::mat4(0.1f), glm::vec3(2.f, 0.f, 0.f));
    glProgramUniformMatrix4fv(main_program, 3, 1, false, value_ptr(model_matrix));
    //
    glProgramUniform3f(main_program, 11, 0.0f, 0.0f, 0.2f); //model_ambient
    glProgramUniform3f(main_program, 12, 0.0f, 0.0f, 0.7f); //_diffuse
    glProgramUniform3f(main_program, 13, 1.f, 1.f, 1.f); //_specular
    glProgramUniform1f(main_program, 14, 0.0f); //shinines
    sphere.draw();

    model_matrix = glm::translate(glm::mat4(0.1f), glm::vec3(-2.f, 0.f, 0.f));
    glProgramUniformMatrix4fv(main_program, 3, 1, false, value_ptr(model_matrix));
    //
    glProgramUniform3f(main_program, 11, 0.2f, 0.0f, 0.0f); //model_ambient
    glProgramUniform3f(main_program, 12, 0.7f, 0.0f, 0.0f); //_diffuse
    glProgramUniform3f(main_program, 13, 1.f, 1.f, 1.f); //_specular
    glProgramUniform1f(main_program, 14, 0.0f); //shinines
    teapot.draw();
}

void Application::render_ui() { const float unit = ImGui::GetFontSize(); }

// ----------------------------------------------------------------------------
// Input Events
// ----------------------------------------------------------------------------

void Application::on_resize(int width, int height) {
    // Calls the default implementation to set the class variables.
    PV112Application::on_resize(width, height);
}

void Application::on_mouse_move(double x, double y) {
    double delta_x = x - prev_mouse_x;
    prev_mouse_x = x;

    if (MLB_is_pressed) current_xz_plane_view_angle += (float)(delta_x/200.0 * 3.14159 / 2.0);
}

void Application::on_mouse_button(int button, int action, int mods) {
    //todo nahodit flag podle skutecneho stavu left-mouse-button (LMB)
    MLB_is_pressed = true;
}

void Application::on_key_pressed(int key, int scancode, int action, int mods) {
    // Calls default implementation that invokes compile_shaders when 'R key is hit.
    PV112Application::on_key_pressed(key, scancode, action, mods);

    if (action == GLFW_PRESS) {
        switch (key) {
        case GLFW_KEY_F:
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            break;
        case GLFW_KEY_L:
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            break;
        case GLFW_KEY_Q:
            // Task 5.2: Map the change between directional and point light to a key on a keyboard.
            break;
        }
    }
}
